@@ -19,10 +19,12 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-var isArray = Array.isArray;
 var domain;
 
+exports.usingDomains = false;
+
 function EventEmitter() {
+  this.domain = null;
   if (exports.usingDomains) {
     // if there is an active domain, then attach to it.
     domain = domain || require('domain');
@@ -30,6 +32,8 @@ function EventEmitter() {
       this.domain = domain.active;
     }
   }
+  this._events = this._events || {};
+  this._maxListeners = this._maxListeners || defaultMaxListeners;
 }
 exports.EventEmitter = EventEmitter;
 
@@ -41,49 +45,47 @@ exports.EventEmitter = EventEmitter;
 // that to be increased. Set to zero for unlimited.
 var defaultMaxListeners = 10;
 EventEmitter.prototype.setMaxListeners = function(n) {
-  if (!this._events) this._events = {};
+  if (typeof n !== 'number' || n < 0)
+    throw TypeError('n must be a positive number');
   this._maxListeners = n;
 };
 
-// non-global reference, for speed.
-var PROCESS;
+EventEmitter.prototype.emit = function(type) {
+  var er, handler, len, args, i, listeners;
+
+  if (!this._events)
+    this._events = {};
 
-EventEmitter.prototype.emit = function() {
-  var type = arguments[0];
   // If there is no 'error' event listener then throw.
   if (type === 'error') {
-    if (!this._events || !this._events.error ||
-        (isArray(this._events.error) && !this._events.error.length))
-    {
+    if (!this._events.error ||
+        (typeof this._events.error === 'object' &&
+         !this._events.error.length)) {
+      er = arguments[1];
       if (this.domain) {
-        var er = arguments[1];
-        er.domain_emitter = this;
+        if (!er) er = new TypeError('Uncaught, unspecified "error" event.');
+        er.domainEmitter = this;
         er.domain = this.domain;
-        er.domain_thrown = false;
+        er.domainThrown = false;
         this.domain.emit('error', er);
-        return false;
-      }
-
-      if (arguments[1] instanceof Error) {
-        throw arguments[1]; // Unhandled 'error' event
+      } else if (er instanceof Error) {
+        throw er; // Unhandled 'error' event
       } else {
-        throw new Error("Uncaught, unspecified 'error' event.");
+        throw TypeError('Uncaught, unspecified "error" event.');
       }
       return false;
     }
   }
 
-  if (!this._events) return false;
-  var handler = this._events[type];
-  if (!handler) return false;
+  handler = this._events[type];
 
-  if (typeof handler == 'function') {
-    if (this.domain) {
-      PROCESS = PROCESS || process;
-      if (this !== PROCESS) {
-        this.domain.enter();
-      }
-    }
+  if (typeof handler === 'undefined')
+    return false;
+
+  if (this.domain && this !== process)
+    this.domain.enter();
+
+  if (typeof handler === 'function') {
     switch (arguments.length) {
       // fast cases
       case 1:
@@ -97,78 +99,58 @@ EventEmitter.prototype.emit = function() {
         break;
       // slower
       default:
-        var l = arguments.length;
-        var args = new Array(l - 1);
-        for (var i = 1; i < l; i++) args[i - 1] = arguments[i];
+        len = arguments.length;
+        args = new Array(len - 1);
+        for (i = 1; i < len; i++)
+          args[i - 1] = arguments[i];
         handler.apply(this, args);
     }
-    if (this.domain && this !== PROCESS) {
-      this.domain.exit();
-    }
-    return true;
+  } else if (typeof handler === 'object') {
+    len = arguments.length;
+    args = new Array(len - 1);
+    for (i = 1; i < len; i++)
+      args[i - 1] = arguments[i];
 
-  } else if (isArray(handler)) {
-    if (this.domain) {
-      PROCESS = PROCESS || process;
-      if (this !== PROCESS) {
-        this.domain.enter();
-      }
-    }
-    var l = arguments.length;
-    var args = new Array(l - 1);
-    for (var i = 1; i < l; i++) args[i - 1] = arguments[i];
-
-    var listeners = handler.slice();
-    for (var i = 0, l = listeners.length; i < l; i++) {
+    listeners = handler.slice();
+    len = listeners.length;
+    for (i = 0; i < len; i++)
       listeners[i].apply(this, args);
-    }
-    if (this.domain && this !== PROCESS) {
-      this.domain.exit();
-    }
-    return true;
-
-  } else {
-    return false;
   }
+
+  if (this.domain && this !== process)
+    this.domain.exit();
+
+  return true;
 };
 
 EventEmitter.prototype.addListener = function(type, listener) {
-  if ('function' !== typeof listener) {
-    throw new Error('addListener only takes instances of Function');
-  }
+  var m;
 
-  if (!this._events) this._events = {};
+  if (typeof listener !== 'function')
+    throw TypeError('listener must be a function');
 
-  // To avoid recursion in the case that type == "newListeners"! Before
-  // adding it to the listeners, first emit "newListeners".
-  if (this._events.newListener) {
+  if (!this._events)
+    this._events = {};
+
+  // To avoid recursion in the case that type === "newListener"! Before
+  // adding it to the listeners, first emit "newListener".
+  if (this._events.newListener)
     this.emit('newListener', type, typeof listener.listener === 'function' ?
               listener.listener : listener);
-  }
 
-  if (!this._events[type]) {
+  if (!this._events[type])
     // Optimize the case of one listener. Don't need the extra array object.
     this._events[type] = listener;
-  } else if (isArray(this._events[type])) {
-
+  else if (typeof this._events[type] === 'object')
     // If we've already got an array, just append.
     this._events[type].push(listener);
-
-  } else {
+  else
     // Adding the second element, need to change to array.
     this._events[type] = [this._events[type], listener];
 
-  }
-
   // Check for listener leak
-  if (isArray(this._events[type]) && !this._events[type].warned) {
-    var m;
-    if (this._maxListeners !== undefined) {
-      m = this._maxListeners;
-    } else {
-      m = defaultMaxListeners;
-    }
-
+  if (typeof this._events[type] === 'object' && !this._events[type].warned) {
+    m = this._maxListeners;
     if (m && m > 0 && this._events[type].length > m) {
       this._events[type].warned = true;
       console.error('(node) warning: possible EventEmitter memory ' +
@@ -185,81 +167,84 @@ EventEmitter.prototype.addListener = function(type, listener) {
 EventEmitter.prototype.on = EventEmitter.prototype.addListener;
 
 EventEmitter.prototype.once = function(type, listener) {
-  if ('function' !== typeof listener) {
-    throw new Error('.once only takes instances of Function');
+  if (typeof listener !== 'function')
+    throw TypeError('listener must be a function');
+
+  function g() {
+    this.removeListener(type, g);
+    listener.apply(this, arguments);
   }
 
-  var self = this;
-  function g() {
-    self.removeListener(type, g);
-    listener.apply(this, arguments);
-  };
-
   g.listener = listener;
-  self.on(type, g);
+  this.on(type, g);
 
   return this;
 };
 
 // emits a 'removeListener' event iff the listener was removed
 EventEmitter.prototype.removeListener = function(type, listener) {
-  if ('function' !== typeof listener) {
-    throw new Error('removeListener only takes instances of Function');
-  }
+  var list, position, length, i;
 
-  // does not use listeners(), so no side effect of creating _events[type]
-  if (!this._events || !this._events[type]) return this;
+  if (typeof listener !== 'function')
+    throw TypeError('listener must be a function');
 
-  var list = this._events[type];
+  if (!this._events || !this._events[type])
+    return this;
 
-  if (isArray(list)) {
-    var position = -1;
-    for (var i = 0, length = list.length; i < length; i++) {
+  list = this._events[type];
+  length = list.length;
+  position = -1;
+
+  if (list === listener ||
+      (typeof list.listener === 'function' && list.listener === listener)) {
+    delete this._events[type];
+    if (this._events.removeListener)
+      this.emit('removeListener', type, listener);
+
+  } else if (typeof list === 'object') {
+    for (i = length; i-- > 0;) {
       if (list[i] === listener ||
-          (list[i].listener && list[i].listener === listener))
-      {
+          (list[i].listener && list[i].listener === listener)) {
         position = i;
         break;
       }
     }
 
-    if (position < 0) return this;
-    list.splice(position, 1);
-    if (list.length == 0)
+    if (position < 0)
+      return this;
+
+    if (list.length === 1) {
+      list.length = 0;
       delete this._events[type];
-
-    if (this._events.removeListener) {
-      this.emit('removeListener', type, listener);
+    } else {
+      list.splice(position, 1);
     }
-  } else if (list === listener ||
-             (list.listener && list.listener === listener))
-  {
-    delete this._events[type];
 
-    if (this._events.removeListener) {
+    if (this._events.removeListener)
       this.emit('removeListener', type, listener);
-    }
   }
 
   return this;
 };
 
 EventEmitter.prototype.removeAllListeners = function(type) {
-  if (!this._events) return this;
+  var key, listeners;
 
-  // fast path
+  if (!this._events)
+    return this;
+
+  // not listening for removeListener, no need to emit
   if (!this._events.removeListener) {
-    if (arguments.length === 0) {
+    if (arguments.length === 0)
       this._events = {};
-    } else if (type && this._events && this._events[type]) {
-      this._events[type] = null;
-    }
+    else if (this._events[type])
+      delete this._events[type];
     return this;
   }
 
-  // slow(ish) path, emit 'removeListener' events for all removals
+  // emit removeListener for all listeners on all events
   if (arguments.length === 0) {
-    for (var key in this._events) {
+    for (key in this._events) {
       if (key === 'removeListener') continue;
       this.removeAllListeners(key);
     }
@@ -268,24 +253,38 @@ EventEmitter.prototype.removeAllListeners = function(type) {
     return this;
   }
 
-  var listeners = this._events[type];
-  if (isArray(listeners)) {
-    while (listeners.length) {
-      // LIFO order
-      this.removeListener(type, listeners[listeners.length - 1]);
-    }
-  } else if (listeners) {
+  listeners = this._events[type];
+
+  if (typeof listeners === 'function') {
     this.removeListener(type, listeners);
+  } else {
+    // LIFO order
+    while (listeners.length)
+      this.removeListener(type, listeners[listeners.length - 1]);
   }
-  this._events[type] = null;
+  delete this._events[type];
 
   return this;
 };
 
 EventEmitter.prototype.listeners = function(type) {
-  if (!this._events || !this._events[type]) return [];
-  if (!isArray(this._events[type])) {
-    return [this._events[type]];
-  }
-  return this._events[type].slice(0);
+  var ret;
+  if (!this._events || !this._events[type])
+    ret = [];
+  else if (typeof this._events[type] === 'function')
+    ret = [this._events[type]];
+  else
+    ret = this._events[type].slice();
+  return ret;
+};
+
+EventEmitter.listenerCount = function(emitter, type) {
+  var ret;
+  if (!emitter._events || !emitter._events[type])
+    ret = 0;
+  else if (typeof emitter._events[type] === 'function')
+    ret = 1;
+  else
+    ret = emitter._events[type].length;
+  return ret;
 };
