@@ -506,10 +506,16 @@ CryptoStream.prototype._read = function read(size) {
     if (this.ondata) {
       this.ondata(pool, start, start + bytesRead);
 
-      // Consume data automatically
-      // simple/test-https-drain fails without it
-      this.push(pool.slice(start, start + bytesRead));
-      this.read(bytesRead);
+      // Deceive streams2
+      var self = this;
+
+      setImmediate(function() {
+        // Force state.reading to set to false
+        self.push('');
+
+        // Try reading more, we most likely have some data
+        self.read(0);
+      });
     } else {
       this.push(pool.slice(start, start + bytesRead));
     }
