@@ -35,9 +35,13 @@
         value: process.constructor
       }
     });
+    EventEmitter.call(process);
 
     process.EventEmitter = EventEmitter; // process.EventEmitter is deprecated
 
+    // do this good and early, since it handles errors.
+    startup.processFatal();
+
     startup.globalVariables();
     startup.globalTimeouts();
     startup.globalConsole();
@@ -45,7 +49,6 @@
     startup.processAssert();
     startup.processConfig();
     startup.processNextTick();
-    startup.processMakeCallback();
     startup.processStdio();
     startup.processKillAndExit();
     startup.processSignalHandlers();
@@ -108,13 +111,12 @@
         // global.v8debug object about a connection, and runMain when
         // that occurs.  --isaacs
 
-        setTimeout(Module.runMain, 50);
+        var debugTimeout = +process.env.NODE_DEBUG_TIMEOUT || 50;
+        setTimeout(Module.runMain, debugTimeout);
 
       } else {
-        // REMOVEME: nextTick should not be necessary. This hack to get
-        // test/simple/test-exception-handler2.js working.
         // Main entry point into most programs:
-        process.nextTick(Module.runMain);
+        Module.runMain();
       }
 
     } else {
@@ -140,7 +142,6 @@
 
       } else {
         // Read all of stdin - execute it.
-        process.stdin.resume();
         process.stdin.setEncoding('utf8');
 
         var code = '';
@@ -162,6 +163,9 @@
     global.GLOBAL = global;
     global.root = global;
     global.Buffer = NativeModule.require('buffer').Buffer;
+    process.binding('buffer').setFastBufferConstructor(global.Buffer);
+    process.domain = null;
+    process._exiting = false;
   };
 
   startup.globalTimeouts = function() {
@@ -212,6 +216,80 @@
     return startup._lazyConstants;
   };
 
+  startup.processFatal = function() {
+    // call into the active domain, or emit uncaughtException,
+    // and exit if there are no listeners.
+    process._fatalException = function(er) {
+      var caught = false;
+      if (process.domain) {
+        var domain = process.domain;
+        var domainModule = NativeModule.require('domain');
+        var domainStack = domainModule._stack;
+
+        // ignore errors on disposed domains.
+        //
+        // XXX This is a bit stupid.  We should probably get rid of
+        // domain.dispose() altogether.  It's almost always a terrible
+        // idea.  --isaacs
+        if (domain._disposed)
+          return true;
+
+        er.domain = domain;
+        er.domainThrown = true;
+        // wrap this in a try/catch so we don't get infinite throwing
+        try {
+          // One of three things will happen here.
+          //
+          // 1. There is a handler, caught = true
+          // 2. There is no handler, caught = false
+          // 3. It throws, caught = false
+          //
+          // If caught is false after this, then there's no need to exit()
+          // the domain, because we're going to crash the process anyway.
+          caught = domain.emit('error', er);
+
+          // Exit all domains on the stack.  Uncaught exceptions end the
+          // current tick and no domains should be left on the stack
+          // between ticks.
+          var domainModule = NativeModule.require('domain');
+          domainStack.length = 0;
+          domainModule.active = process.domain = null;
+        } catch (er2) {
+          // The domain error handler threw!  oh no!
+          // See if another domain can catch THIS error,
+          // or else crash on the original one.
+          // If the user already exited it, then don't double-exit.
+          if (domain === domainModule.active)
+            domainStack.pop();
+          if (domainStack.length) {
+            var parentDomain = domainStack[domainStack.length - 1];
+            process.domain = domainModule.active = parentDomain;
+            caught = process._fatalException(er2);
+          } else
+            caught = false;
+        }
+      } else {
+        caught = process.emit('uncaughtException', er);
+      }
+      // if someone handled it, then great.  otherwise, die in C++ land
+      // since that means that we'll exit the process, emit the 'exit' event
+      if (!caught) {
+        try {
+          if (!process._exiting) {
+            process._exiting = true;
+            process.emit('exit', 1);
+          }
+        } catch (er) {
+          // nothing to be done about it at this point.
+        }
+      }
+      // if we handled an error, then make sure any ticks get processed
+      if (caught)
+        process._needTickCallback();
+      return caught;
+    };
+  };
+
   var assert;
   startup.processAssert = function() {
     // Note that calls to assert() are pre-processed out by JS2C for the
@@ -237,29 +315,28 @@
     });
   };
 
-  startup.processMakeCallback = function() {
-    process._makeCallback = function(obj, fn, args) {
-      var domain = obj.domain;
-      if (domain) {
-        if (domain._disposed) return;
-        domain.enter();
-      }
-
-      var ret = fn.apply(obj, args);
-
-      if (domain) domain.exit();
-
-      // process the nextTicks after each time we get called.
-      process._tickCallback();
-      return ret;
-    };
-  };
-
   startup.processNextTick = function() {
+    var _needTickCallback = process._needTickCallback;
     var nextTickQueue = [];
-    var nextTickIndex = 0;
+    var needSpinner = true;
     var inTick = false;
-    var tickDepth = 0;
+
+    // this infobox thing is used so that the C++ code in src/node.cc
+    // can have easy accesss to our nextTick state, and avoid unnecessary
+    // calls into process._tickCallback.
+    // order is [length, index, depth]
+    // Never write code like this without very good reason!
+    var infoBox = process._tickInfoBox;
+    var length = 0;
+    var index = 1;
+    var depth = 2;
+
+    process.nextTick = nextTick;
+    // needs to be accessible from cc land
+    process._nextDomainTick = _nextDomainTick;
+    process._tickCallback = _tickCallback;
+    process._tickDomainCallback = _tickDomainCallback;
+    process._tickFromSpinner = _tickFromSpinner;
 
     // the maximum number of times it'll process something like
     // nextTick(function f(){nextTick(f)})
@@ -271,16 +348,83 @@
     process.maxTickDepth = 1000;
 
     function tickDone(tickDepth_) {
-      tickDepth = tickDepth_ || 0;
-      nextTickQueue.splice(0, nextTickIndex);
-      nextTickIndex = 0;
-      inTick = false;
-      if (nextTickQueue.length) {
-        process._needTickCallback();
+      if (infoBox[length] !== 0) {
+        if (infoBox[length] <= infoBox[index]) {
+          nextTickQueue = [];
+          infoBox[length] = 0;
+        } else {
+          nextTickQueue.splice(0, infoBox[index]);
+          infoBox[length] = nextTickQueue.length;
+          if (needSpinner) {
+            _needTickCallback();
+            needSpinner = false;
+          }
+        }
       }
+      inTick = false;
+      infoBox[index] = 0;
+      infoBox[depth] = tickDepth_;
     }
 
-    process._tickCallback = function(fromSpinner) {
+    function maxTickWarn() {
+      // XXX Remove all this maxTickDepth stuff in 0.11
+      var msg = '(node) warning: Recursive process.nextTick detected. ' +
+                'This will break in the next version of node. ' +
+                'Please use setImmediate for recursive deferral.';
+      if (process.throwDeprecation)
+        throw new Error(msg);
+      else if (process.traceDeprecation)
+        console.trace(msg);
+      else
+        console.error(msg);
+    }
+
+    function _tickFromSpinner() {
+      needSpinner = true;
+      // coming from spinner, reset!
+      if (infoBox[depth] !== 0)
+        infoBox[depth] = 0;
+      // no callbacks to run
+      if (infoBox[length] === 0)
+        return infoBox[index] = infoBox[depth] = 0;
+      process._tickCallback();
+    }
+
+    // run callbacks that have no domain
+    // using domains will cause this to be overridden
+    function _tickCallback() {
+      var callback, nextTickLength, threw;
+
+      if (inTick) return;
+      if (infoBox[length] === 0) {
+        infoBox[index] = 0;
+        infoBox[depth] = 0;
+        return;
+      }
+      inTick = true;
+
+      while (infoBox[depth]++ < process.maxTickDepth) {
+        nextTickLength = infoBox[length];
+        if (infoBox[index] === nextTickLength)
+          return tickDone(0);
+
+        while (infoBox[index] < nextTickLength) {
+          callback = nextTickQueue[infoBox[index]++].callback;
+          threw = true;
+          try {
+            callback();
+            threw = false;
+          } finally {
+            if (threw) tickDone(infoBox[depth]);
+          }
+        }
+      }
+
+      tickDone(0);
+    }
+
+    function _tickDomainCallback() {
+      var nextTickLength, tock, callback, threw;
 
       // if you add a nextTick in a domain's error handler, then
       // it's possible to cycle indefinitely.  Normally, the tickDone
@@ -288,73 +432,81 @@
       // that error handler ALSO triggers multiple MakeCallbacks, then
       // it'll try to keep clearing the queue, since the finally block
       // fires *before* the error hits the top level and is handled.
-      if (tickDepth >= process.maxTickDepth) {
-        if (fromSpinner) {
-          // coming in from the event queue.  reset.
-          tickDepth = 0;
-        } else {
-          if (nextTickQueue.length) {
-            process._needTickCallback();
-          }
-          return;
-        }
-      }
-
-      if (!nextTickQueue.length) return tickDone();
+      if (infoBox[depth] >= process.maxTickDepth)
+        return _needTickCallback();
 
       if (inTick) return;
       inTick = true;
 
       // always do this at least once.  otherwise if process.maxTickDepth
       // is set to some negative value, or if there were repeated errors
-      // preventing tickDepth from being cleared, we'd never process any
+      // preventing depth from being cleared, we'd never process any
       // of them.
-      do {
-        tickDepth++;
-        var nextTickLength = nextTickQueue.length;
-        if (nextTickLength === 0) return tickDone();
-        while (nextTickIndex < nextTickLength) {
-          var tock = nextTickQueue[nextTickIndex++];
-          var callback = tock.callback;
+      while (infoBox[depth]++ < process.maxTickDepth) {
+        nextTickLength = infoBox[length];
+        if (infoBox[index] === nextTickLength)
+          return tickDone(0);
+
+        while (infoBox[index] < nextTickLength) {
+          tock = nextTickQueue[infoBox[index]++];
+          callback = tock.callback;
           if (tock.domain) {
             if (tock.domain._disposed) continue;
             tock.domain.enter();
           }
-          var threw = true;
+          threw = true;
           try {
             callback();
             threw = false;
           } finally {
             // finally blocks fire before the error hits the top level,
-            // so we can't clear the tickDepth at this point.
-            if (threw) tickDone(tickDepth);
+            // so we can't clear the depth at this point.
+            if (threw) tickDone(infoBox[depth]);
           }
           if (tock.domain) {
             tock.domain.exit();
           }
         }
-        nextTickQueue.splice(0, nextTickIndex);
-        nextTickIndex = 0;
-
-        // continue until the max depth or we run out of tocks.
-      } while (tickDepth < process.maxTickDepth &&
-               nextTickQueue.length > 0);
-
-      tickDone();
-    };
-
-    process.nextTick = function(callback) {
-      // on the way out, don't bother.
-      // it won't get fired anyway.
-      if (process._exiting) return;
-
-      var tock = { callback: callback };
-      if (process.domain) tock.domain = process.domain;
-      nextTickQueue.push(tock);
-      if (nextTickQueue.length) {
-        process._needTickCallback();
       }
-    };
+
+      tickDone(0);
+    }
+
+    function nextTick(callback) {
+      // on the way out, don't bother. it won't get fired anyway.
+      if (process._exiting)
+        return;
+      if (infoBox[depth] >= process.maxTickDepth)
+        maxTickWarn();
+
+      var obj = { callback: callback, domain: null };
+
+      nextTickQueue.push(obj);
+      infoBox[length]++;
+
+      if (needSpinner) {
+        _needTickCallback();
+        needSpinner = false;
+      }
+    }
+
+    function _nextDomainTick(callback) {
+      // on the way out, don't bother. it won't get fired anyway.
+      if (process._exiting)
+        return;
+      if (infoBox[depth] >= process.maxTickDepth)
+        maxTickWarn();
+
+      var obj = { callback: callback, domain: process.domain };
+
+      nextTickQueue.push(obj);
+      infoBox[length]++;
+
+      if (needSpinner) {
+        _needTickCallback();
+        needSpinner = false;
+      }
+    }
   };
 
   function evalScript(name) {
@@ -417,14 +569,20 @@
         break;
 
       case 'PIPE':
+      case 'TCP':
         var net = NativeModule.require('net');
-        stream = new net.Stream(fd);
+        stream = new net.Socket({
+          fd: fd,
+          readable: false,
+          writable: true
+        });
 
-        // FIXME Should probably have an option in net.Stream to create a
+        // FIXME Should probably have an option in net.Socket to create a
         // stream from an existing fd which is writable only. But for now
         // we'll just add this hack and set the `readable` member to false.
         // Test: ./node test/fixtures/echo.js < /etc/passwd
         stream.readable = false;
+        stream.read = null;
         stream._type = 'pipe';
 
         // FIXME Hack to have stream not keep the event loop alive.
@@ -484,18 +642,26 @@
       switch (tty_wrap.guessHandleType(fd)) {
         case 'TTY':
           var tty = NativeModule.require('tty');
-          stdin = new tty.ReadStream(fd);
+          stdin = new tty.ReadStream(fd, {
+            highWaterMark: 0,
+            readable: true,
+            writable: false
+          });
           break;
 
         case 'FILE':
           var fs = NativeModule.require('fs');
-          stdin = new fs.ReadStream(null, {fd: fd});
+          stdin = new fs.ReadStream(null, { fd: fd });
           break;
 
         case 'PIPE':
+        case 'TCP':
           var net = NativeModule.require('net');
-          stdin = new net.Stream(fd);
-          stdin.readable = true;
+          stdin = new net.Socket({
+            fd: fd,
+            readable: true,
+            writable: false
+          });
           break;
 
         default:
@@ -507,16 +673,23 @@
       stdin.fd = fd;
 
       // stdin starts out life in a paused state, but node doesn't
-      // know yet.  Call pause() explicitly to unref() it.
-      stdin.pause();
+      // know yet.  Explicitly to readStop() it to put it in the
+      // not-reading state.
+      if (stdin._handle && stdin._handle.readStop) {
+        stdin._handle.reading = false;
+        stdin._readableState.reading = false;
+        stdin._handle.readStop();
+      }
 
-      // when piping stdin to a destination stream,
-      // let the data begin to flow.
-      var pipe = stdin.pipe;
-      stdin.pipe = function(dest, opts) {
-        stdin.resume();
-        return pipe.call(stdin, dest, opts);
-      };
+      // if the user calls stdin.pause(), then we need to stop reading
+      // immediately, so that the process can close down.
+      stdin.on('pause', function() {
+        if (!stdin._handle)
+          return;
+        stdin._readableState.reading = false;
+        stdin._handle.reading = false;
+        stdin._handle.readStop();
+      });
 
       return stdin;
     });
@@ -552,7 +725,7 @@
       }
 
       if (r) {
-        throw errnoException(errno, 'kill');
+        throw errnoException(process._errno, 'kill');
       }
 
       return true;
@@ -560,10 +733,6 @@
   };
 
   startup.processSignalHandlers = function() {
-    // Not supported on Windows.
-    if (process.platform === 'win32')
-      return;
-
     // Load events module in order to access prototype elements on process like
     // process.addListener.
     var signalWraps = {};
@@ -590,7 +759,7 @@
         var r = wrap.start(signum);
         if (r) {
           wrap.close();
-          throw errnoException(errno, 'uv_signal_start');
+          throw errnoException(process._errno, 'uv_signal_start');
         }
 
         signalWraps[type] = wrap;
@@ -688,8 +857,8 @@
 
     var nativeModule = new NativeModule(id);
 
-    nativeModule.compile();
     nativeModule.cache();
+    nativeModule.compile();
 
     return nativeModule.exports;
   };
