@@ -19,24 +19,35 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+
+
+
+process.env.NODE_DEBUGGER_TIMEOUT = 2000;
 var common = require('../common');
 var assert = require('assert');
 var debug = require('_debugger');
 
+var debugPort = common.PORT + 1337;
+debug.port = debugPort;
 var spawn = require('child_process').spawn;
 
+setTimeout(function() {
+  if (nodeProcess) nodeProcess.kill('SIGTERM');
+  throw new Error('timeout');
+}, 10000).unref();
+
 
 var resCount = 0;
 var p = new debug.Protocol();
-p.onResponse = function (res) {
+p.onResponse = function(res) {
   resCount++;
 };
 
-p.execute("Type: connect\r\n" +
-          "V8-Version: 3.0.4.1\r\n" +
-          "Protocol-Version: 1\r\n" +
-          "Embedding-Host: node v0.3.3-pre\r\n" +
-          "Content-Length: 0\r\n\r\n");
+p.execute('Type: connect\r\n' +
+          'V8-Version: 3.0.4.1\r\n' +
+          'Protocol-Version: 1\r\n' +
+          'Embedding-Host: node v0.3.3-pre\r\n' +
+          'Content-Length: 0\r\n\r\n');
 assert.equal(1, resCount);
 
 // Make sure split messages go in.
@@ -53,14 +64,15 @@ parts.push('{"seq":12,"type":"event","event":"break","body":' +
 assert.equal(78, parts[2].length);
 bodyLength += parts[2].length;
 
-parts.push('.[anonymous](req=#<an IncomingMessage>, res=#<a ServerResponse>)",' +
-           '"sourceLine"');
+parts.push('.[anonymous](req=#<an IncomingMessage>, ' +
+           'res=#<a ServerResponse>)","sourceLine"');
 assert.equal(78, parts[3].length);
 bodyLength += parts[3].length;
 
-parts.push(':45,"sourceColumn":4,"sourceLineText":"    debugger;","script":' +
-           '{"id":24,"name":"/home/ryan/projects/node/benchmark/http_simple.js",' +
-           '"lineOffset":0,"columnOffset":0,"lineCount":98}}}');
+parts.push(':45,"sourceColumn":4,"sourceLineText":"    debugger;",' +
+           '"script":{"id":24,"name":"/home/ryan/projects/node/' +
+           'benchmark/http_simple.js","lineOffset":0,"columnOffset":0,' +
+           '"lineCount":98}}}');
 assert.equal(180, parts[4].length);
 bodyLength += parts[4].length;
 
@@ -77,10 +89,12 @@ assert.equal(2, resCount);
 var d = 'Content-Length: 466\r\n\r\n' +
         '{"seq":10,"type":"event","event":"afterCompile","success":true,' +
         '"body":{"script":{"handle":1,"type":"script","name":"dns.js",' +
-        '"id":34,"lineOffset":0,"columnOffset":0,"lineCount":241,"sourceStart":' +
-        '"(function (module, exports, require) {var dns = process.binding(\'cares\')' +
+        '"id":34,"lineOffset":0,"columnOffset":0,"lineCount":241,' +
+        '"sourceStart":"(function (module, exports, require) {' +
+        'var dns = process.binding(\'cares\')' +
         ';\\nvar ne","sourceLength":6137,"scriptType":2,"compilationType":0,' +
-        '"context":{"ref":0},"text":"dns.js (lines: 241)"}},"refs":[{"handle":0' +
+        '"context":{"ref":0},"text":"dns.js (lines: 241)"}},"refs":' +
+        '[{"handle":0' +
         ',"type":"context","text":"#<a ContextMirror>"}],"running":true}' +
         'Content-Length: 119\r\n\r\n' +
         '{"seq":11,"type":"event","event":"scriptCollected","success":true,' +
@@ -90,24 +104,26 @@ assert.equal(4, resCount);
 
 var expectedConnections = 0;
 var tests = [];
-function addTest (cb) {
+function addTest(cb) {
   expectedConnections++;
   tests.push(cb);
 }
 
-addTest(function (client, done) {
-  console.error("requesting version");
-  client.reqVersion(function (v) {
-    console.log("version: %s", v);
+addTest(function(client, done) {
+  console.error('requesting version');
+  client.reqVersion(function(err, v) {
+    assert.ok(!err);
+    console.log('version: %s', v);
     assert.equal(process.versions.v8, v);
     done();
   });
 });
 
-addTest(function (client, done) {
-  console.error("requesting scripts");
-  client.reqScripts(function () {
-    console.error("got %d scripts", Object.keys(client.scripts).length);
+addTest(function(client, done) {
+  console.error('requesting scripts');
+  client.reqScripts(function(err) {
+    assert.ok(!err);
+    console.error('got %d scripts', Object.keys(client.scripts).length);
 
     var foundMainScript = false;
     for (var k in client.scripts) {
@@ -122,63 +138,81 @@ addTest(function (client, done) {
   });
 });
 
-addTest(function (client, done) {
-  console.error("eval 2+2");
-  client.reqEval("2+2", function (res) {
-    assert.ok(res.success);
+addTest(function(client, done) {
+  console.error('eval 2+2');
+  client.reqEval('2+2', function(err, res) {
     console.error(res);
-    assert.equal('4', res.body.text);
-    assert.equal(4, res.body.value);
+    assert.ok(!err);
+    assert.equal('4', res.text);
+    assert.equal(4, res.value);
     done();
   });
 });
 
 
 var connectCount = 0;
+var script = 'setTimeout(function () { console.log("blah"); });' +
+             'setInterval(function () {}, 1000000);';
+
+var nodeProcess;
 
 function doTest(cb, done) {
-  var nodeProcess = spawn(process.execPath,
-      ['-e', 'setInterval(function () { console.log("blah"); }, 100);']);
+  var args = ['--debug=' + debugPort, '-e', script];
+  nodeProcess = spawn(process.execPath, args);
 
-  nodeProcess.stdout.once('data', function () {
-    console.log(">>> new node process: %d", nodeProcess.pid);
-    process.kill(nodeProcess.pid, "SIGUSR1");
-    console.log(">>> signaling it with SIGUSR1");
+  nodeProcess.stdout.once('data', function(c) {
+    console.log('>>> new node process: %d', nodeProcess.pid);
+    var failed = true;
+    try {
+      process._debugProcess(nodeProcess.pid);
+      failed = false;
+    } finally {
+      // At least TRY not to leave zombie procs if this fails.
+      if (failed)
+        nodeProcess.kill('SIGTERM');
+    }
+    console.log('>>> starting debugger session');
   });
 
   var didTryConnect = false;
   nodeProcess.stderr.setEncoding('utf8');
-  var b = ''
-  nodeProcess.stderr.on('data', function (data) {
+  var b = '';
+  nodeProcess.stderr.on('data', function(data) {
+    console.error('got stderr data %j', data);
+    nodeProcess.stderr.resume();
     b += data;
-    if (didTryConnect == false && /debugger listening on port/.test(b)) {
+    if (didTryConnect == false &&
+        b.match(/debugger listening on port/)) {
       didTryConnect = true;
 
       setTimeout(function() {
         // Wait for some data before trying to connect
         var c = new debug.Client();
-        process.stdout.write(">>> connecting...");
-        c.connect(debug.port)
-        c.on('ready', function () {
+        console.error('>>> connecting...');
+        c.connect(debug.port);
+        c.on('break', function(brk) {
+          c.reqContinue(function() {});
+        });
+        c.on('ready', function() {
           connectCount++;
-          console.log("ready!");
-          cb(c, function () {
-            console.error(">>> killing node process %d\n\n", nodeProcess.pid);
+          console.log('ready!');
+          cb(c, function() {
+            console.error('>>> killing node process %d\n\n', nodeProcess.pid);
             nodeProcess.kill();
             done();
           });
         });
-      }, 100);
+      });
     }
   });
 }
 
 
-function run () {
+function run() {
   var t = tests[0];
   if (!t) return;
 
-  doTest(t, function () {
+  doTest(t, function() {
     tests.shift();
     run();
   });
@@ -186,7 +220,8 @@ function run () {
 
 run();
 
-process.on('exit', function() {
-  assert.equal(expectedConnections, connectCount);
+process.on('exit', function(code) {
+  if (!code)
+    assert.equal(expectedConnections, connectCount);
 });
 
