@@ -20,8 +20,8 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 if (!process.features.tls_npn) {
-  console.error("Skipping because node compiled without OpenSSL or " +
-                "with old OpenSSL version.");
+  console.error('Skipping because node compiled without OpenSSL or ' +
+                'with old OpenSSL version.');
   process.exit(0);
 }
 
@@ -31,7 +31,7 @@ var common = require('../common'),
     tls = require('tls');
 
 function filenamePEM(n) {
-  return require('path').join(common.fixturesDir, 'keys', n + ".pem");
+  return require('path').join(common.fixturesDir, 'keys', n + '.pem');
 }
 
 function loadPEM(n) {
@@ -45,25 +45,37 @@ var serverOptions = {
   NPNProtocols: ['a', 'b', 'c']
 };
 
-var clientsOptions = [{
-  key: serverOptions.key,
-  cert: serverOptions.cert,
-  crl: serverOptions.crl,
-  NPNProtocols: ['a', 'b', 'c']
-},{
-  key: serverOptions.key,
-  cert: serverOptions.cert,
-  crl: serverOptions.crl,
-  NPNProtocols: ['c', 'b', 'e']
-},{
-  key: serverOptions.key,
-  cert: serverOptions.cert,
-  crl: serverOptions.crl,
-  NPNProtocols: ['first-priority-unsupported', 'x', 'y']
-}];
-
 var serverPort = common.PORT;
 
+var clientsOptions = [{
+  port: serverPort,
+  key: serverOptions.key,
+  cert: serverOptions.cert,
+  crl: serverOptions.crl,
+  NPNProtocols: ['a', 'b', 'c'],
+  rejectUnauthorized: false
+},{
+  port: serverPort,
+  key: serverOptions.key,
+  cert: serverOptions.cert,
+  crl: serverOptions.crl,
+  NPNProtocols: ['c', 'b', 'e'],
+  rejectUnauthorized: false
+},{
+  port: serverPort,
+  key: serverOptions.key,
+  cert: serverOptions.cert,
+  crl: serverOptions.crl,
+  rejectUnauthorized: false
+},{
+  port: serverPort,
+  key: serverOptions.key,
+  cert: serverOptions.cert,
+  crl: serverOptions.crl,
+  NPNProtocols: ['first-priority-unsupported', 'x', 'y'],
+  rejectUnauthorized: false
+}];
+
 var serverResults = [],
     clientsResults = [];
 
@@ -74,7 +86,7 @@ server.listen(serverPort, startTest);
 
 function startTest() {
   function connectClient(options, callback) {
-    var client = tls.connect(serverPort, 'localhost', options, function() {
+    var client = tls.connect(options, function() {
       clientsResults.push(client.npnProtocol);
       client.destroy();
 
@@ -85,15 +97,19 @@ function startTest() {
   connectClient(clientsOptions[0], function() {
     connectClient(clientsOptions[1], function() {
       connectClient(clientsOptions[2], function() {
-        server.close();
+        connectClient(clientsOptions[3], function() {
+          server.close();
+        });
       });
     });
   });
-};
+}
 
 process.on('exit', function() {
   assert.equal(serverResults[0], clientsResults[0]);
   assert.equal(serverResults[1], clientsResults[1]);
-  assert.equal(serverResults[2], 'first-priority-unsupported');
+  assert.equal(serverResults[2], 'http/1.1');
   assert.equal(clientsResults[2], false);
+  assert.equal(serverResults[3], 'first-priority-unsupported');
+  assert.equal(clientsResults[3], false);
 });
